@@ -1,7 +1,7 @@
 use crate::auditwheel::PlatformTag;
 use crate::build_context::{BridgeModel, ProjectLayout};
 use crate::cross_compile::{find_sysconfigdata, parse_sysconfigdata};
-use crate::python_interpreter::InterpreterKind;
+use crate::python_interpreter::{InterpreterKind, MINIMUM_PYTHON_MINOR};
 use crate::BuildContext;
 use crate::CargoToml;
 use crate::Metadata21;
@@ -17,8 +17,32 @@ use std::env;
 use std::io;
 use std::path::PathBuf;
 
+// This is used for BridgeModel::Bindings("pyo3-ffi") and BridgeModel::Bindings("pyo3").
+// These should be treated almost identically but must be correctly identified
+// as one or the other in logs. pyo3-ffi is ordered first because it is newer
+// and more restrictive.
+const PYO3_BINDING_CRATES: [&str; 2] = ["pyo3-ffi", "pyo3"];
+
+fn pyo3_minimum_python_minor_version(major_version: u64, minor_version: u64) -> Option<usize> {
+    if (major_version, minor_version) >= (0, 16) {
+        Some(7)
+    } else if (major_version, minor_version) >= (0, 12) {
+        Some(6)
+    } else {
+        None
+    }
+}
+
+fn pyo3_ffi_minimum_python_minor_version(major_version: u64, minor_version: u64) -> Option<usize> {
+    if (major_version, minor_version) >= (0, 16) {
+        pyo3_minimum_python_minor_version(major_version, minor_version)
+    } else {
+        None
+    }
+}
+
 /// High level API for building wheels from a crate which is also used for the CLI
-#[derive(Debug, Serialize, Deserialize, clap::Parser, Clone, Eq, PartialEq)]
+#[derive(Debug, Default, Serialize, Deserialize, clap::Parser, Clone, Eq, PartialEq)]
 #[serde(default)]
 pub struct BuildOptions {
     /// Control the platform tag on linux.
@@ -40,23 +64,29 @@ pub struct BuildOptions {
         parse(try_from_str)
     )]
     pub platform_tag: Option<PlatformTag>,
-    #[clap(short, long)]
+
     /// The python versions to build wheels for, given as the names of the
     /// interpreters. Uses autodiscovery if not explicitly set.
-    pub interpreter: Option<Vec<PathBuf>>,
+    #[clap(short, long)]
+    pub interpreter: Vec<PathBuf>,
+
     /// Which kind of bindings to use. Possible values are pyo3, rust-cpython, cffi and bin
     #[clap(short, long)]
     pub bindings: Option<String>,
-    #[clap(short = 'm', long = "manifest-path", parse(from_os_str), name = "PATH")]
+
     /// The path to the Cargo.toml
+    #[clap(short = 'm', long = "manifest-path", parse(from_os_str), name = "PATH")]
     pub manifest_path: Option<PathBuf>,
+
     /// The directory to store the built wheels in. Defaults to a new "wheels"
     /// directory in the project's target directory
     #[clap(short, long, parse(from_os_str))]
     pub out: Option<PathBuf>,
+
     /// Don't check for manylinux compliance
     #[clap(long = "skip-auditwheel")]
     pub skip_auditwheel: bool,
+
     /// For manylinux targets, use zig to ensure compliance for the chosen manylinux version
     ///
     /// Default to manylinux2010/manylinux_2_12 if you do not specify an `--compatibility`
@@ -64,9 +94,11 @@ pub struct BuildOptions {
     /// Make sure you installed zig with `pip install maturin[zig]`
     #[clap(long)]
     pub zig: bool,
+
     /// The --target option for cargo
     #[clap(long, name = "TRIPLE", env = "CARGO_BUILD_TARGET")]
     pub target: Option<String>,
+
     /// Extra arguments that will be passed to cargo as `cargo rustc [...] [arg1] [arg2] -- [...]`
     ///
     /// Use as `--cargo-extra-args="--my-arg"`
@@ -76,35 +108,19 @@ pub struct BuildOptions {
     /// be a bit flaky.
     #[clap(long = "cargo-extra-args")]
     pub cargo_extra_args: Vec<String>,
+
     /// Extra arguments that will be passed to rustc as `cargo rustc [...] -- [...] [arg1] [arg2]`
     ///
     /// Use as `--rustc-extra-args="--my-arg"`
     #[clap(long = "rustc-extra-args")]
     pub rustc_extra_args: Vec<String>,
+
     /// Control whether to build universal2 wheel for macOS or not.
     /// Only applies to macOS targets, do nothing otherwise.
     #[clap(long)]
     pub universal2: bool,
 }
 
-impl Default for BuildOptions {
-    fn default() -> Self {
-        BuildOptions {
-            platform_tag: None,
-            interpreter: Some(vec![]),
-            bindings: None,
-            manifest_path: None,
-            out: None,
-            skip_auditwheel: false,
-            zig: false,
-            target: None,
-            cargo_extra_args: Vec::new(),
-            rustc_extra_args: Vec::new(),
-            universal2: false,
-        }
-    }
-}
-
 impl BuildOptions {
     /// Get cargo manifest file path
     fn manifest_path(&self) -> Result<PathBuf> {
@@ -116,18 +132,21 @@ impl BuildOptions {
         let current_dir = env::current_dir()
             .context("Failed to detect current directory ಠ_ಠ")?
             .canonicalize()?;
-        let pyproject = PyProjectToml::new(&current_dir).context("pyproject.toml is invalid")?;
-        if let Some(path) = pyproject.manifest_path() {
-            println!("🔗 Found cargo manifest path in pyproject.toml");
-            // pyproject.toml must be placed at top directory
-            let manifest_dir = path
-                .parent()
-                .context("missing parent directory")?
-                .canonicalize()?;
-            if !manifest_dir.starts_with(&current_dir) {
-                bail!("Cargo.toml can not be placed outside of the directory containing pyproject.toml");
+        if current_dir.join("pyproject.toml").is_file() {
+            let pyproject =
+                PyProjectToml::new(&current_dir).context("pyproject.toml is invalid")?;
+            if let Some(path) = pyproject.manifest_path() {
+                println!("🔗 Found cargo manifest path in pyproject.toml");
+                // pyproject.toml must be placed at top directory
+                let manifest_dir = path
+                    .parent()
+                    .context("missing parent directory")?
+                    .canonicalize()?;
+                if !manifest_dir.starts_with(&current_dir) {
+                    bail!("Cargo.toml can not be placed outside of the directory containing pyproject.toml");
+                }
+                return Ok(path.to_path_buf());
             }
-            return Ok(path.to_path_buf());
         }
         // check Cargo.toml in current directory
         let path = PathBuf::from("Cargo.toml");
@@ -206,9 +225,6 @@ impl BuildOptions {
             }
         }
         cargo_extra_args = split_extra_args(&cargo_extra_args)?;
-        if let Some(ref target) = self.target {
-            cargo_extra_args.extend(vec!["--target".to_string(), target.clone()]);
-        }
 
         let cargo_metadata_extra_args = extract_cargo_metadata_args(&cargo_extra_args)?;
 
@@ -256,13 +272,12 @@ impl BuildOptions {
             None => PathBuf::from(&cargo_metadata.target_directory).join("wheels"),
         };
 
-        let interpreter = match self.interpreter {
-            // Only build a source distribution
-            Some(ref interpreter) if interpreter.is_empty() => vec![],
-            // User given list of interpreters
-            Some(interpreter) => find_interpreter(&bridge, &interpreter, &target, None)?,
+        let interpreter = if self.interpreter.is_empty() {
             // Auto-detect interpreters
-            None => find_interpreter(&bridge, &[], &target, get_min_python_minor(&metadata21))?,
+            find_interpreter(&bridge, &[], &target, get_min_python_minor(&metadata21))?
+        } else {
+            // User given list of interpreters
+            find_interpreter(&bridge, &self.interpreter, &target, None)?
         };
 
         let mut rustc_extra_args = self.rustc_extra_args.clone();
@@ -316,7 +331,12 @@ impl BuildOptions {
                     Some(target.get_minimum_manylinux_tag())
                 }
             } else {
-                None
+                // Defaults to musllinux_1_2 for musl target if it's not bin bindings
+                if target.is_musl_target() && !matches!(bridge, BridgeModel::Bin) {
+                    Some(PlatformTag::Musllinux { x: 1, y: 2 })
+                } else {
+                    None
+                }
             });
         if platform_tag == Some(PlatformTag::manylinux1()) {
             eprintln!("⚠️  Warning: manylinux1 is unsupported by the Rust compiler.");
@@ -329,6 +349,13 @@ impl BuildOptions {
             );
         }
 
+        let target_dir = cargo_extra_args
+            .iter()
+            .position(|x| x == "--target-dir")
+            .and_then(|i| cargo_extra_args.get(i + 1))
+            .map(PathBuf::from)
+            .unwrap_or_else(|| cargo_metadata.target_directory.clone().into_std_path_buf());
+
         Ok(BuildContext {
             target,
             bridge,
@@ -337,6 +364,7 @@ impl BuildOptions {
             crate_name: crate_name.to_string(),
             module_name,
             manifest_path: manifest_file,
+            target_dir,
             out: wheel_dir,
             release,
             strip,
@@ -382,44 +410,44 @@ fn has_abi3(cargo_metadata: &Metadata) -> Result<Option<(u8, u8)>> {
         .resolve
         .as_ref()
         .context("Expected cargo to return metadata with resolve")?;
-    let pyo3_packages = resolve
-        .nodes
-        .iter()
-        .filter(|package| cargo_metadata[&package.id].name == "pyo3")
-        .collect::<Vec<_>>();
-    match pyo3_packages.as_slice() {
-        [pyo3_crate] => {
-            // Find the minimal abi3 python version. If there is none, abi3 hasn't been selected
-            // This parser abi3-py{major}{minor} and returns the minimal (major, minor) tuple
-            let abi3_selected = pyo3_crate.features.iter().any(|x| x == "abi3");
+    for &lib in PYO3_BINDING_CRATES.iter() {
+        let pyo3_packages = resolve
+            .nodes
+            .iter()
+            .filter(|package| cargo_metadata[&package.id].name.as_str() == lib)
+            .collect::<Vec<_>>();
+        match pyo3_packages.as_slice() {
+            [pyo3_crate] => {
+                // Find the minimal abi3 python version. If there is none, abi3 hasn't been selected
+                // This parser abi3-py{major}{minor} and returns the minimal (major, minor) tuple
+                let abi3_selected = pyo3_crate.features.iter().any(|x| x == "abi3");
 
-            let min_abi3_version = pyo3_crate
-                .features
-                .iter()
-                .filter(|x| x.starts_with("abi3-py") && x.len() >= "abi3-pyxx".len())
-                .map(|x| {
-                    Ok((
-                        (x.as_bytes()[7] as char).to_string().parse::<u8>()?,
-                        x[8..].parse::<u8>()?,
-                    ))
-                })
-                .collect::<Result<Vec<(u8, u8)>>>()
-                .context("Bogus pyo3 cargo features")?
-                .into_iter()
-                .min();
-            if abi3_selected && min_abi3_version.is_none() {
-                bail!(
-                    "You have selected the `abi3` feature but not a minimum version (e.g. the `abi3-py36` feature). \
-                    maturin needs a minimum version feature to build abi3 wheels."
-                )
+                let min_abi3_version = pyo3_crate
+                    .features
+                    .iter()
+                    .filter(|x| x.starts_with("abi3-py") && x.len() >= "abi3-pyxx".len())
+                    .map(|x| {
+                        Ok((
+                            (x.as_bytes()[7] as char).to_string().parse::<u8>()?,
+                            x[8..].parse::<u8>()?,
+                        ))
+                    })
+                    .collect::<Result<Vec<(u8, u8)>>>()
+                    .context(format!("Bogus {} cargo features", lib))?
+                    .into_iter()
+                    .min();
+                if abi3_selected && min_abi3_version.is_none() {
+                    bail!(
+                        "You have selected the `abi3` feature but not a minimum version (e.g. the `abi3-py36` feature). \
+                        maturin needs a minimum version feature to build abi3 wheels."
+                    )
+                }
+                return Ok(min_abi3_version);
             }
-            Ok(min_abi3_version)
+            _ => continue,
         }
-        _ => bail!(format!(
-            "Expected exactly one pyo3 dependency, found {}",
-            pyo3_packages.len()
-        )),
     }
+    Ok(None)
 }
 
 /// Tries to determine the [BridgeModel] for the target crate
@@ -434,6 +462,18 @@ pub fn find_bridge(cargo_metadata: &Metadata, bridge: Option<&str>) -> Result<Br
         .iter()
         .map(|node| (cargo_metadata[&node.id].name.as_ref(), node))
         .collect();
+    let packages: HashMap<&str, &cargo_metadata::Package> = cargo_metadata
+        .packages
+        .iter()
+        .filter_map(|pkg| {
+            let name = &pkg.name;
+            if name == "pyo3" || name == "pyo3-ffi" || name == "cpython" {
+                Some((name.as_ref(), pkg))
+            } else {
+                None
+            }
+        })
+        .collect();
 
     let bridge = if let Some(bindings) = bridge {
         if bindings == "cffi" {
@@ -449,20 +489,27 @@ pub fn find_bridge(cargo_metadata: &Metadata, bridge: Option<&str>) -> Result<Br
                 );
             }
 
-            BridgeModel::Bindings(bindings.to_string())
+            BridgeModel::Bindings(bindings.to_string(), MINIMUM_PYTHON_MINOR)
         }
     } else if deps.get("pyo3").is_some() {
-        BridgeModel::Bindings("pyo3".to_string())
+        let ver = &packages["pyo3"].version;
+        let minor =
+            pyo3_minimum_python_minor_version(ver.major, ver.minor).unwrap_or(MINIMUM_PYTHON_MINOR);
+        BridgeModel::Bindings("pyo3".to_string(), minor)
+    } else if deps.get("pyo3-ffi").is_some() {
+        let ver = &packages["pyo3-ffi"].version;
+        let minor = pyo3_ffi_minimum_python_minor_version(ver.major, ver.minor)
+            .unwrap_or(MINIMUM_PYTHON_MINOR);
+        BridgeModel::Bindings("pyo3-ffi".to_string(), minor)
     } else if deps.contains_key("cpython") {
         println!("🔗 Found rust-cpython bindings");
-        BridgeModel::Bindings("rust_cpython".to_string())
+        BridgeModel::Bindings("rust_cpython".to_string(), MINIMUM_PYTHON_MINOR)
     } else {
         let package = &cargo_metadata[resolve.root.as_ref().unwrap()];
         let targets: Vec<_> = package
             .targets
             .iter()
-            .map(|target| target.crate_types.iter())
-            .flatten()
+            .flat_map(|target| target.crate_types.iter())
             .map(String::as_str)
             .collect();
 
@@ -475,28 +522,30 @@ pub fn find_bridge(cargo_metadata: &Metadata, bridge: Option<&str>) -> Result<Br
         }
     };
 
-    if BridgeModel::Bindings("pyo3".to_string()) == bridge {
-        let pyo3_node = deps["pyo3"];
-        if !pyo3_node.features.contains(&"extension-module".to_string()) {
-            let version = cargo_metadata[&pyo3_node.id].version.to_string();
-            println!(
-                "⚠️  Warning: You're building a library without activating pyo3's \
-                 `extension-module` feature. \
-                 See https://pyo3.rs/v{}/building_and_distribution.html#linking",
-                version
-            );
-        }
+    for &lib in PYO3_BINDING_CRATES.iter() {
+        if bridge.is_bindings(lib) {
+            let pyo3_node = deps[lib];
+            if !pyo3_node.features.contains(&"extension-module".to_string()) {
+                let version = cargo_metadata[&pyo3_node.id].version.to_string();
+                println!(
+                    "⚠️  Warning: You're building a library without activating {}'s \
+                     `extension-module` feature. \
+                     See https://pyo3.rs/v{}/building_and_distribution.html#linking",
+                    lib, version
+                );
+            }
 
-        return if let Some((major, minor)) = has_abi3(cargo_metadata)? {
-            println!(
-                "🔗 Found pyo3 bindings with abi3 support for Python ≥ {}.{}",
-                major, minor
-            );
-            Ok(BridgeModel::BindingsAbi3(major, minor))
-        } else {
-            println!("🔗 Found pyo3 bindings");
-            Ok(bridge)
-        };
+            return if let Some((major, minor)) = has_abi3(cargo_metadata)? {
+                println!(
+                    "🔗 Found {} bindings with abi3 support for Python ≥ {}.{}",
+                    lib, major, minor
+                );
+                Ok(BridgeModel::BindingsAbi3(major, minor))
+            } else {
+                println!("🔗 Found {} bindings", lib);
+                Ok(bridge)
+            };
+        }
     }
 
     Ok(bridge)
@@ -538,7 +587,7 @@ pub fn find_interpreter(
     min_python_minor: Option<usize>,
 ) -> Result<Vec<PythonInterpreter>> {
     match bridge {
-        BridgeModel::Bindings(binding_name) => {
+        BridgeModel::Bindings(binding_name, _) => {
             let mut interpreter = if !interpreter.is_empty() {
                 PythonInterpreter::check_executables(interpreter, target, bridge)
                     .context("The given list of python interpreters is invalid")?
@@ -555,7 +604,7 @@ pub fn find_interpreter(
                 }
             }
 
-            if binding_name == "pyo3" && target.is_unix() && target.cross_compiling() {
+            if binding_name.starts_with("pyo3") && target.is_unix() && target.cross_compiling() {
                 if let Some(cross_lib_dir) = std::env::var_os("PYO3_CROSS_LIB_DIR") {
                     println!("⚠️ Cross-compiling is poorly supported");
                     let host_python = &interpreter[0];
@@ -645,7 +694,8 @@ pub fn find_interpreter(
                 PythonInterpreter::check_executables(interpreter, target, bridge)
                     .unwrap_or_default()
             } else {
-                PythonInterpreter::find_all(target, bridge, min_python_minor).unwrap_or_default()
+                PythonInterpreter::find_all(target, bridge, Some(*minor as usize))
+                    .unwrap_or_default()
             };
             // Ideally, we wouldn't want to use any python interpreter without abi3 at all.
             // Unfortunately, on windows we need one to figure out base_prefix for a linker
@@ -759,11 +809,11 @@ mod test {
 
         assert!(matches!(
             find_bridge(&pyo3_mixed, None),
-            Ok(BridgeModel::Bindings(_))
+            Ok(BridgeModel::Bindings(..))
         ));
         assert!(matches!(
             find_bridge(&pyo3_mixed, Some("pyo3")),
-            Ok(BridgeModel::Bindings(_))
+            Ok(BridgeModel::Bindings(..))
         ));
 
         assert!(find_bridge(&pyo3_mixed, Some("rust-cpython")).is_err());
@@ -778,11 +828,11 @@ mod test {
 
         assert!(matches!(
             find_bridge(&pyo3_pure, None),
-            Ok(BridgeModel::BindingsAbi3(3, 6))
+            Ok(BridgeModel::BindingsAbi3(3, 7))
         ));
         assert!(matches!(
             find_bridge(&pyo3_pure, Some("pyo3")),
-            Ok(BridgeModel::BindingsAbi3(3, 6))
+            Ok(BridgeModel::BindingsAbi3(3, 7))
         ));
         assert!(find_bridge(&pyo3_pure, Some("rust-cpython")).is_err());
     }
@@ -804,7 +854,7 @@ mod test {
 
         assert!(matches!(
             find_bridge(&pyo3_pure, None).unwrap(),
-            BridgeModel::Bindings(_)
+            BridgeModel::Bindings(..)
         ));
     }
 
