@@ -12,11 +12,11 @@ use fs_err::File;
 use sha2::{Digest, Sha256};
 use std::collections::HashMap;
 use std::ffi::OsStr;
-#[cfg(not(target_os = "windows"))]
+#[cfg(target_family = "unix")]
 use std::fs::OpenOptions;
 use std::io;
 use std::io::{Read, Write};
-#[cfg(not(target_os = "windows"))]
+#[cfg(target_family = "unix")]
 use std::os::unix::fs::OpenOptionsExt;
 use std::path::{Path, PathBuf};
 use std::process::{Command, Output};
@@ -164,7 +164,7 @@ impl ModuleWriter for PathWriter {
 
         // We only need to set the executable bit on unix
         let mut file = {
-            #[cfg(not(target_os = "windows"))]
+            #[cfg(target_family = "unix")]
             {
                 OpenOptions::new()
                     .create(true)
@@ -214,7 +214,7 @@ impl ModuleWriter for WheelWriter {
         permissions: u32,
     ) -> Result<()> {
         // The zip standard mandates using unix style paths
-        let target = target.as_ref().to_str().unwrap().replace("\\", "/");
+        let target = target.as_ref().to_str().unwrap().replace('\\', "/");
 
         // Unlike users which can use the develop subcommand, the tests have to go through
         // packing a zip which pip than has to unpack. This makes this 2-3 times faster
@@ -298,7 +298,7 @@ impl WheelWriter {
             zip::CompressionMethod::Deflated
         };
         let options = zip::write::FileOptions::default().compression_method(compression_method);
-        let record_filename = self.record_file.to_str().unwrap().replace("\\", "/");
+        let record_filename = self.record_file.to_str().unwrap().replace('\\', "/");
         self.zip.start_file(&record_filename, options)?;
         for (filename, hash, len) in self.record {
             self.zip
@@ -441,8 +441,8 @@ where
 
 /// Checks if user has provided their own header at `target/header.h`, otherwise
 /// we run cbindgen to generate one.
-fn cffi_header(crate_dir: &Path, tempdir: &TempDir) -> Result<PathBuf> {
-    let maybe_header = crate_dir.join("target").join("header.h");
+fn cffi_header(crate_dir: &Path, target_dir: &Path, tempdir: &TempDir) -> Result<PathBuf> {
+    let maybe_header = target_dir.join("header.h");
 
     if maybe_header.is_file() {
         println!("💼 Using the existing header at {}", maybe_header.display());
@@ -486,9 +486,13 @@ fn cffi_header(crate_dir: &Path, tempdir: &TempDir) -> Result<PathBuf> {
 /// how to load the shared library without the header and then writes those instructions to a
 /// file called `ffi.py`. This `ffi.py` will expose an object called `ffi`. This object is used
 /// in `__init__.py` to load the shared library into a module called `lib`.
-pub fn generate_cffi_declarations(crate_dir: &Path, python: &Path) -> Result<String> {
+pub fn generate_cffi_declarations(
+    crate_dir: &Path,
+    target_dir: &Path,
+    python: &Path,
+) -> Result<String> {
     let tempdir = tempdir()?;
-    let header = cffi_header(crate_dir, &tempdir)?;
+    let header = cffi_header(crate_dir, target_dir, &tempdir)?;
 
     let ffi_py = tempdir.as_ref().join("ffi.py");
 
@@ -622,7 +626,7 @@ pub fn write_bindings_module(
             ..
         } => {
             if !editable {
-                write_python_part(writer, python_module, &module_name)
+                write_python_part(writer, python_module)
                     .context("Failed to add the python module to the package")?;
             }
 
@@ -679,12 +683,13 @@ pub fn write_cffi_module(
     writer: &mut impl ModuleWriter,
     project_layout: &ProjectLayout,
     crate_dir: &Path,
+    target_dir: &Path,
     module_name: &str,
     artifact: &Path,
     python: &Path,
     editable: bool,
 ) -> Result<()> {
-    let cffi_declarations = generate_cffi_declarations(crate_dir, python)?;
+    let cffi_declarations = generate_cffi_declarations(crate_dir, target_dir, python)?;
 
     let module;
 
@@ -695,7 +700,7 @@ pub fn write_cffi_module(
             ref extension_name,
         } => {
             if !editable {
-                write_python_part(writer, python_module, &module_name)
+                write_python_part(writer, python_module)
                     .context("Failed to add the python module to the package")?;
             }
 
@@ -767,24 +772,15 @@ pub fn write_bin(
 }
 
 /// Adds the python part of a mixed project to the writer,
-/// excluding older versions of the native library or generated cffi declarations
 pub fn write_python_part(
     writer: &mut impl ModuleWriter,
     python_module: impl AsRef<Path>,
-    module_name: impl AsRef<Path>,
 ) -> Result<()> {
     use ignore::WalkBuilder;
 
     for absolute in WalkBuilder::new(&python_module).hidden(false).build() {
         let absolute = absolute?.into_path();
-
         let relative = absolute.strip_prefix(python_module.as_ref().parent().unwrap())?;
-
-        // Ignore the cffi folder from develop, if any
-        if relative.starts_with(module_name.as_ref().join(&module_name)) {
-            continue;
-        }
-
         if absolute.is_dir() {
             writer.add_directory(relative)?;
         } else {
@@ -837,5 +833,16 @@ pub fn write_dist_info(
         )?;
     }
 
+    if !metadata21.license_files.is_empty() {
+        let license_files_dir = dist_info_dir.join("license_files");
+        writer.add_directory(&license_files_dir)?;
+        for path in &metadata21.license_files {
+            let filename = path.file_name().with_context(|| {
+                format!("missing file name for license file {}", path.display())
+            })?;
+            writer.add_file(license_files_dir.join(filename), path)?;
+        }
+    }
+
     Ok(())
 }
